/******************************************************************************
 * File: arch.c 
 * 
 * Copyright (C) 2001 Angus Ainslie <aasoft@netscape.net>
 * 
 * This is a PalmPix Archimage.pdb database decoder
 * 
 *       This program is free software; you can redistribute it and/or modify
 *       it for use in non-commercial products. Any commercial use of this 
 *       code or code derived from this code requires my permission.
 * 
 *       This program is distributed in the hope that it will be useful,
 *       but WITHOUT ANY WARRANTY; without even the implied warranty of
 *       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 *****************************************************************************/

#include <stdio.h>

#ifdef __GNUC__
#include <unistd.h>
#else
#include <io.h>
#include <winsock.h>
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <string.h>
#include <fcntl.h>

#include "palm.h"
#include "palmpix.h"
#include "wlut.h"
#include "blut.h"

int verbose = 0;

void ShowHdr( DatabaseHdrType *dHdr );
void GetDirEntries( RecordEntryType *rEntry, dirEntry_t *dEntry, int num, int size );
void ShowPalmPixHeader( pixHdr_t pixHdr );
void ShowPalmPixDB( FILE *inFP, dirEntry_t *dEntry, int num );
void Bias( float bias, int width, int height, UInt8 *data );
void Interpolate( pixHdr_t pixHdr, UInt8 *red, UInt8 *greenR, UInt8 *greenB, UInt8 *blue, RGB_back_t *pp );
void DecodePalmPixDB( FILE *inFP, dirEntry_t *dEntry, int num );
void ShowRecordEntries( RecordEntryType *rEntry, int num );

void ShowHdr( DatabaseHdrType *dHdr )
{
   fprintf( stderr, "name: %s\n", dHdr->name );
   fprintf( stderr, "attr: 0x%x version 0x%x\n", dHdr->attributes, dHdr->version );
   fprintf( stderr, "creationDate:     0x%x\n", dHdr->creationDate );
   fprintf( stderr, "modificationDate: 0x%x\n", dHdr-> modificationDate );
   fprintf( stderr, "LastBackupDate:   0x%x\n", dHdr->lastBackupDate );
   fprintf( stderr, "modficationNumber:0x%x\n", dHdr->modificationNumber );
   fprintf( stderr, "appInfoID:        0x%x\n", dHdr->appInfoID );
   fprintf( stderr, "sortInfoID:       0x%x\n", dHdr->sortInfoID );
   fprintf( stderr, "type:             %c%c%c%c\n", dHdr->type, dHdr->type>>8, dHdr->type>>16, dHdr->type>>24 );
   fprintf( stderr, "creator:          %c%c%c%c\n", dHdr->creator, dHdr->creator>>8, dHdr->creator>>16, dHdr->creator>>24 );
   fprintf( stderr, "uniqueIDSeed:     0x%x\n", dHdr->uniqueIDSeed );
}

void GetDirEntries( RecordEntryType *rEntry, dirEntry_t *dEntry, int num, int size )
{
   int i;
   
   for( i=0; i<num; i++ )
     {
	dEntry[i].offset = ntohl( rEntry[i].localChunkID );
	if( i != num - 1 )
	  dEntry[i].size = ntohl( rEntry[i+1].localChunkID ) - ntohl( rEntry[i].localChunkID );
	else
	  dEntry[i].size = size - ntohl( rEntry[i].localChunkID );
     }
   
}

void ShowPalmPixHeader( pixHdr_t pixHdr )
{
   fprintf( stderr, "numRec:     %d\n", pixHdr.numRec );
   fprintf( stderr, "s0:         %d\n", pixHdr.s0 );		// Always zero
   fprintf( stderr, "date:       %2.2d\\%2.2d\\%2.2d%2.2d\n", pixHdr.month, pixHdr.day, 
	   pixHdr.cent, pixHdr.year );
   fprintf( stderr, "time:       %d:%2.2d:%2.2d\n", pixHdr.hour, pixHdr.min, pixHdr.sec );
   fprintf( stderr, "resolution: %d\n", pixHdr.resolution );	// hires 1, lowres 2
   fprintf( stderr, "width:      %d\n", pixHdr.w );
   fprintf( stderr, "height:     %d\n", pixHdr.h );
   fprintf( stderr, "Thumb size: %d\n", pixHdr.thumbLen );	//Size of the thumbnail following the name
   fprintf( stderr, "GR size:    %d\n", pixHdr.GRsize );
   fprintf( stderr, "R size:     %d\n", pixHdr.Rsize );
   fprintf( stderr, "B size:     %d\n", pixHdr.Bsize );
   fprintf( stderr, "GB size:    %d\n", pixHdr.GBsize );
   fprintf( stderr, "unknown:    %d\n", pixHdr.num );
   fprintf( stderr, "zoom:       %d\n", pixHdr.zoom );

}

void ShowPalmPixDB( FILE *inFP, dirEntry_t *dEntry, int num )
{
   int i;
   pixHdr_t pixHdr;
   char pixName[33];
   
   for( i=0; i<num; i++ )
     {
	if( dEntry[i].size == 196 )
	  {
	     fseek( inFP, dEntry[i].offset, SEEK_SET );
	     fread( &pixHdr, sizeof( pixHdr ), 1, inFP );
	     ShowPalmPixHeader( pixHdr );
	  }
	if( dEntry[i].size == 32 )
	  {
	     fseek( inFP, dEntry[i].offset, SEEK_SET );
	     fread( pixName, 32, 1, inFP );
	     pixName[32] = '\0';
	     fprintf( stderr, "PixName:     %s\n", pixName );
	  }
     }
}

/****************************************************************
 * Bias
 * 
 * Bias is based on the Fast Alternative to Perlin's Bias algorithm 
 * in Graphics Gems IV by
 * Christophe Schlick schlick@labri.u-bordeuax.fr
 ****************************************************************/

void Bias( float bias, int width, int height, UInt8 *data )
{
   int i;
   float num, denom, t;
   
   for( i=0; i<width*height; i++ )
     {
	t = (float)data[i]/256.0;
	num = t;
	denom = (1.0/bias - 2) * (1.0 - t) + 1;
	data[i] = num/denom * 256.0;
     }
}

void Interpolate( pixHdr_t pixHdr, UInt8 *red, UInt8 *greenR, UInt8 *greenB, UInt8 *blue, RGB_back_t *pp )
{
   int idx, offset, ppOff;
   int x, y, rowOff;
   UInt8 r, g, b;
   int rawWidth = pixHdr.w/2;
   
   for( y=1; y<pixHdr.h-1; y++ )
     {
	idx = y/2;
	offset = idx * rawWidth;
	ppOff = y * pixHdr.w;
	
	if( y%2 == 1 )
	  {
	     for( x=1; x<rawWidth-1; x++ )
	       {
		  rowOff = x*2;

		  r = (red[offset+x-1] + red[offset + x] + red[offset + rawWidth + x -1] + red[offset + rawWidth + x])>>2;
		  g = (greenR[offset+x] + greenR[offset+x+rawWidth] + greenB[offset + x - 1] + greenB[offset + x])>>2;
		  b = blue[offset + x];
		  pp[ppOff + rowOff].r = r;
		  pp[ppOff + rowOff].g = g;
		  pp[ppOff + rowOff].b = b;
		
		  r = (red[offset + x] + red[offset + rawWidth + x])>>1;
		  g = greenB[offset + x];
		  b = (blue[offset + x] + blue[offset + x + 1])>>1;
		  pp[ppOff + rowOff + 1].r = r;
		  pp[ppOff + rowOff + 1].g = g;
		  pp[ppOff + rowOff + 1].b = b;
	       }
	  }
	else
	  {
	     for( x=1; x<rawWidth-1; x++ )
	       {
		  rowOff = x*2;
		  
		  r = (red[offset + x - 1] + red[offset + x])>>1;
		  g = greenR[offset + x];
		  b = (blue[offset-rawWidth+x] + blue[offset+x])>>1;
		  pp[ppOff + rowOff].r = r;
		  pp[ppOff + rowOff].g = g;
		  pp[ppOff + rowOff].b = b;

		  r = red[offset+x];
		  g = (greenR[offset+x] + greenR[offset+x+1] + greenB[offset-rawWidth+x] + greenB[offset + x])>>2;
 		  b = (blue[offset+x-rawWidth] + blue[offset+x-rawWidth-1] + blue[offset+x] + blue[offset+x+1])>>2;
		  pp[ppOff + rowOff + 1].r = r;
		  pp[ppOff + rowOff + 1].g = g;
		  pp[ppOff + rowOff + 1].b = b;

	       }
	  }
     }
}

void DecodePalmPixDB( FILE *inFP, dirEntry_t *dEntry, int num )
{
//   int r, i, j, k, lutFd, outFd, width, height, rawWidth, rawHeight;
   int r, i, j, k, outFd, width, height, rawWidth, rawHeight;
   UInt8 *red, *greenR, *greenB, *blue;
   RGB_back_t *pp;
   BitmapFileHeader_t bfHdr;
   BitmapInfoHeader_t biHdr;
   pixHdr_t pixHdr;
   char pixName[37];
   char inRow[400], outRow[400], lastRow[400];
   short lastWord = 0, offset = 0, tmpOffset;
   long rowOff;
   float bias = 0.85;
   
   for( r=0; r<num; r++ )
     {
	if( dEntry[r].size == 196 )
	  {
	     fseek( inFP, dEntry[r+1].offset, SEEK_SET );
	     fread( pixName, 32, 1, inFP );
	     pixName[32] = '\0';
	     
	     strcat( pixName, ".bmp" );
	     
	     fprintf( stderr, "PixName:     %s\n", pixName );
	     
	     outFd = open( pixName, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP );
	     fseek( inFP, dEntry[r].offset, SEEK_SET );
	     fread( &pixHdr, sizeof( pixHdr ), 1, inFP );
	     fprintf( stderr, "pixhdr size: %d\n", sizeof( pixHdr ));

	     width = pixHdr.w;
	     height = pixHdr.h;
	     rawWidth = width/2;
	     rawHeight = height/2;
   
	     red = malloc( rawWidth * rawHeight );

	     if( !red )
	       {
		  fprintf( stderr, "red buffer allocation error" );
		  exit( EXIT_FAILURE );
	       }

	     greenR = malloc( rawWidth * rawHeight );

	     if( !greenR )
	       {
		  fprintf( stderr, "greeen red buffer allocation error" );
		  exit( EXIT_FAILURE );
	       }

	     greenB = malloc( rawWidth * rawHeight );
	     
	     if( !greenB )
	       {
		  fprintf( stderr, "green blue buffer allocation error" );
		  exit( EXIT_FAILURE );
	       }


	     blue = malloc( rawWidth * rawHeight );
	    
	     if( !blue )
	       {
		  fprintf( stderr, "blue buffer allocation error" );
		  exit( EXIT_FAILURE );
	       }


	     memset( red, 0, rawWidth * rawHeight );
	     memset( greenR, 0, rawWidth * rawHeight );
	     memset( greenB, 0, rawWidth * rawHeight );
	     memset( blue, 0, rawWidth * rawHeight );
	     
	     pp = malloc( pixHdr.w * pixHdr.h * sizeof( RGB_t ));
	     
	     memset( pp, 0, pixHdr.w * pixHdr.h * sizeof( RGB_t ));
	     
	     if( !pp )
	       {
		  fprintf( stderr, "pixmap allocation error" );
		  exit( EXIT_FAILURE );
	       }

	     bfHdr.type = 0x4D42;
	     bfHdr.size = sizeof( bfHdr ) + sizeof( biHdr ) + sizeof( RGB_t ) * pixHdr.w * pixHdr.h;     
	     bfHdr.reserved1 = 0;
	     bfHdr.reserved2 = 0;
	     bfHdr.offBits = sizeof( bfHdr ) + sizeof( biHdr );
	     write( outFd, &bfHdr, sizeof( bfHdr ));
	     
	     biHdr.size = sizeof( biHdr );
	     biHdr.width = pixHdr.w;
	     biHdr.height = pixHdr.h;
	     biHdr.planes = 1;
	     biHdr.bitCount = 24;
	     biHdr.compression = BI_RGB;
	     biHdr.sizeImage = 0;
	     biHdr.XpelsPerMeter = 2000;
	     biHdr.YpelsPerMeter = 2000;
	     biHdr.clrUsed = 0;
	     biHdr.clrImportant = 0;
	     write( outFd, &biHdr, sizeof( biHdr ));

	     for( k=0; k<4; k++ )
	       {
		  offset = 0;
		  lastWord = 0;
		  tmpOffset = 0;
		  fseek( inFP, dEntry[r+(4+k*(pixHdr.numRec/4))].offset, SEEK_SET );
		  fread( lastRow, pixHdr.w/2, 1, inFP );
		  fread( inRow, pixHdr.w/2, 1, inFP );
		  
		  switch( k )
		    {
		     case 0:
		       memcpy( greenR, lastRow, rawWidth );
		       break;
		     case 1:
		       memcpy( red, lastRow, rawWidth );
		       break;
		     case 2:
		       memcpy( blue, lastRow, rawWidth );
		       break;
		     case 3:
		       memcpy( greenB, lastRow, rawWidth );
		       break;
		    }

		  rowOff = rawWidth;
		  
		  for( j=1; j<rawHeight; j++ )
		    {
		       DecodeRow( inRow, lastRow, outRow, &tmpOffset, &lastWord, wlut, blut, rawWidth );
		       memcpy( lastRow, outRow, rawWidth );
		       
		       switch( k )
			 {
			  case 0:
			    memcpy( &greenR[rowOff], lastRow, rawWidth );
			    break;
			  case 1:
			    memcpy( &red[rowOff], lastRow, rawWidth );
			    break;
			  case 2:
			    memcpy( &blue[rowOff], lastRow, rawWidth );
			    break;
			    
			  case 3:
			    memcpy( &greenB[rowOff], lastRow, rawWidth );
			    break;
			 }
		       
		       rowOff += rawWidth;
		       offset += tmpOffset;
		       fseek( inFP, dEntry[r+(4+k*(pixHdr.numRec/4))].offset + offset + rawWidth, SEEK_SET );
		       fread( inRow, rawWidth, 1, inFP );
		    }
	       }
	     
	     Bias( bias, rawWidth, rawHeight, greenR );
	     Bias( bias, rawWidth, rawHeight, red );
	     Bias( bias, rawWidth, rawHeight, blue );
	     Bias( bias, rawWidth, rawHeight, greenB );
	     
	     Interpolate( pixHdr, red, greenR, greenB, blue, pp );

	     for( i=pixHdr.h; i>0; i-- )
	       write( outFd, &pp[i * pixHdr.w], pixHdr.w * sizeof( RGB_t ));
	     
	     free( red );
	     free( greenR );
	     free( greenB );
	     free( blue );
	     
	     free( pp );
	     
	     close( outFd );
	  }
     }
   
}

void ShowRecordEntries( RecordEntryType *rEntry, int num )
{
   int i, size;
   
   for( i=0; i<num; i++ )
     {
	fprintf( stderr, "Record %d\n", i );
	fprintf( stderr, "localChunkID:   0x%x\n", ntohl( rEntry[i].localChunkID ));
	fprintf( stderr, "attr:           0x%x\n", rEntry[i].attributes );
	fprintf( stderr, "UniqueID:       0x%x%x%x\n", rEntry[i].uniqueID[0], rEntry[i].uniqueID[1], rEntry[i].uniqueID[2] );
	if( i != num-1 )
	  {
	     size = ntohl( rEntry[i+1].localChunkID ) - ntohl( rEntry[i].localChunkID );
	     fprintf( stderr, "size:           %d\n", size );
	  }
     }
}

int main( int argc, char **argv )
{
   int inFd, recs, size;
   DatabaseHdrType pHdr;
   RecordEntryType *rEntry;
   dirEntry_t *dEntry;
   int opt;
   char inFile[128];
   FILE *fP;

   if( argc < 2 )
     {
#ifdef __GNUC__
	fprintf( stderr, "Not enough args:\n%s [-v] -i input_file\n\t-v verbose\n", argv[0] );
#else
	fprintf( stderr, "Not enough args:\n%s [-v] input_file\n\t-v verbose\n", argv[0] );
#endif
	exit( EXIT_FAILURE );
     }
   
#ifdef __GNUC__
   while( -1 != ( opt = getopt( argc, argv, "i:o:v" )))
     {
	switch( opt )
	  {
	   case 'v': 
	     verbose++;
	     break;
	   
	   case 'i': 
	     strcpy( inFile, optarg );
	     break;
	  }
     }

   inFd = open( inFile, O_RDONLY );
   fP = fdopen( inFd, "rw" );
#else
   inFd = open( argv[1], O_RDONLY | O_BINARY );
   strcpy( outFile, argv[2] );
   fP = fdopen( inFd, "rw" );
#endif
   
   if( inFd == -1 )
     {
	perror( "File open" );
	exit( EXIT_FAILURE );
     }

   if( fP == NULL )
     {
	perror( "Stream Open" );
	exit( EXIT_FAILURE );
     }
   
   if( verbose )
     fprintf( stderr, "Input File: %s\n", inFile );

   if( sizeof( pHdr ) != read( inFd, &pHdr, sizeof( pHdr )))
     {
	perror( "Header Read" );
	exit( EXIT_FAILURE );
     }

   recs = ntohs( pHdr.recordList.numRecords );

   rEntry = (RecordEntryType *)malloc( sizeof( RecordEntryType ) * recs );
   if( rEntry == NULL )
     {
	perror( "Record list malloc" );
	exit( EXIT_FAILURE );
     }

   dEntry = (dirEntry_t *)malloc( sizeof( dirEntry_t ) * recs );
   if( rEntry == NULL )
     {
	perror( "Directory list malloc" );
	exit( EXIT_FAILURE );
     }

   if( sizeof( RecordEntryType ) * recs  != read( inFd, rEntry, sizeof( RecordEntryType ) * recs))
     {
	perror( "Record List Read" );
	exit( EXIT_FAILURE );
     }
   
   fseek( fP, SEEK_END, 0 );
   size = ftell( fP );
   fseek( fP, SEEK_SET, 0 );
   GetDirEntries( rEntry, dEntry, recs, size );
   ShowPalmPixDB( fP, dEntry, recs );
   DecodePalmPixDB( fP, dEntry, recs );

   return( 0 );
}
   
