/******************************************************************************
 * File: decode.c 
 * 
 * Copyright (C) 2001 Angus Ainslie <aasoft@netscape.net>
 * 
 * This is a part of a PalmPix Archimage.pdb database decoder
 * 
 *       This program is free software; you can redistribute it and/or modify
 *       it for use in non-commercial products. Any commercial use of this 
 *       code or code derived from this code requires my permission.
 * 
 *       This program is distributed in the hope that it will be useful,
 *       but WITHOUT ANY WARRANTY; without even the implied warranty of
 *       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 *****************************************************************************/

#include <inttypes.h>
#include <stdio.h>
#include <sys/socket.h>

#include "palm.h"
#include "palmpix.h"

void DecodeRow( uint8_t *compData, uint8_t *lastRow, uint8_t *unCompData, uint16_t *offset, short *firstWord, uint16_t *PPLutsW, uint8_t *PPLuts, uint16_t halfWidth );

void DecodeRow( uint8_t *compData, uint8_t *lastRow, uint8_t *unCompData, uint16_t *offset, short *firstWord, uint16_t *PPLutsW, uint8_t *PPLuts, uint16_t halfWidth )
{
   uint8_t *saveStartP, tmpB7;
   int16_t tmpW3, tmpResult, tmpW0, idx, resultW;
   uint32_t lutIdx, tmpL0, tmpL2, tmpL5;
   
   saveStartP = compData;

   tmpW3 = 32 - *firstWord;

   tmpResult = 32 - tmpW3;
   
   tmpL5 = compData[0]<<24 | compData[1]<<16 | compData[2]<<8 | compData[3] ;
   tmpL5 = tmpL5<<tmpResult;
   tmpL0 = tmpL5;
   
   compData += 4;
   
   tmpL0 = tmpL0 >> 24;
   
   unCompData[0] = (uint8_t)tmpL0;
   
   tmpL5 = tmpL5 << 8;
   tmpW3 = tmpW3 - 8;
   idx = 1;           
   
   while( idx < halfWidth )
     {
	if( tmpW3 < 12 )
	  {
	     tmpW0 = 16 - tmpW3;
	     tmpL2 = ( compData[0]<<8 | compData[1] ) << tmpW0;
	     tmpL5 = tmpL5 | tmpL2;
	     tmpW3 += 16;
	     compData += 2;
	  }

	lutIdx = tmpL5 >> 20;

	tmpB7 = PPLuts[lutIdx];
	
	tmpL5 = tmpL5 << tmpB7;
	
	tmpW3 = tmpW3 - tmpB7;
	
	resultW = ( unCompData[idx-1] + lastRow[idx] ) >> 1;
	
	resultW = resultW + PPLutsW[lutIdx];
	
	if( resultW > 255 )
	  resultW = 255;
	
	if( resultW < 0 ) 
	  resultW = 0;
	
	unCompData[idx++] = (uint8_t)resultW;
	  
     }
   
   while( tmpW3 > 0 )
     {
	compData--;
	tmpW3 -= 8;
     }
     
   *offset = compData - saveStartP;
   *firstWord = -tmpW3;

}
